import os
import socket
import select


class BaseEventFD(object):
    """Class implementing event objects that has a fd that can be selected.

    This EventFD class implements the same functions as a regular Event but it
    has a file descriptor. The file descriptor can be accessed using the fileno function.
    This event can be passed to select, poll and it will block until the event will be set.

    Reference: https://github.com/palaviv/eventfd/blob/develop/eventfd/_eventfd.py
    """

    _DATA = None

    def __init__(self):
        self._flag = False
        self._read_fd = None
        self._write_fd = None

    def _read(self, len):
        return os.read(self._read_fd, len)

    def _write(self, data):
        os.write(self._write_fd, data)

    def is_set(self):
        """Return true if and only if the internal flag is true."""
        return self._flag

    def clear(self):
        """Reset the internal flag to false.

        Subsequently, threads calling wait() will block until set() is called to
        set the internal flag to true again.

        """
        if self._flag:
            self._flag = False
            assert self._read(len(self._DATA)) == self._DATA

    def unsafe_write(self):
        self._write(self._DATA)

    def unsafe_read(self):
        self._read(len(self._DATA))

    def set(self):
        """Set the internal flag to true.

        All threads waiting for it to become true are awakened. Threads
        that call wait() once the flag is true will not block at all.

        """
        if not self._flag:
            self._flag = True
            self._write(self._DATA)

    def wait(self, timeout=None):
        """Block until the internal flag is true.

        If the internal flag is true on entry, return immediately. Otherwise,
        block until another thread calls set() to set the flag to true, or until
        the optional timeout occurs.

        When the timeout argument is present and not None, it should be a
        floating point number specifying a timeout for the operation in seconds
        (or fractions thereof).

        This method returns the internal flag on exit, so it will always return
        True except if a timeout is given and the operation times out.

        """
        if not self._flag:
            ret = select.select([self], [], [], timeout)
            assert ret[0] in [[self], []]
        return self._flag

    def fileno(self):
        """Return a file descriptor that can be selected.

        You should not use this directly pass the EventFD object instead.

        Reference: https://docs.python.org/3/library/select.html#select.select
        """
        return self._read_fd

    def __del__(self):
        """Closes the file descriptors"""
        raise NotImplementedError


class PipeEventFD(BaseEventFD):

    _DATA = b"A"

    def __init__(self):
        super(PipeEventFD, self).__init__()
        self._read_fd, self._write_fd = os.pipe()

    def __del__(self):
        os.close(self._read_fd)
        os.close(self._write_fd)


class SocketEventFD(BaseEventFD):

    _DATA = b'A'

    def __init__(self):
        super(SocketEventFD, self).__init__()
        temp_fd = socket.socket()
        temp_fd.bind(("127.0.0.1", 0))
        temp_fd.listen(1)
        self._read_fd = socket.create_connection(temp_fd.getsockname())
        self._write_fd, _ = temp_fd.accept()
        temp_fd.close()

    def _read(self, len):
        return self._read_fd.recv(len)

    def _write(self, data):
        self._write_fd.send(data)

    def fileno(self):
        return self._read_fd.fileno()

    def __del__(self):
        self._read_fd.close()
        self._write_fd.close()


def eventfd():
    if os.name != "nt":         # Linux
        return PipeEventFD()
    else:                       # Windows
        return SocketEventFD()



# setup.py
from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()
# install_requires = (this_directory / 'requirements.txt').read_text().splitlines()

version = '0.0.39'

config = {
    'name': 'py-netty',
    'url': 'https://github.com/ruanhao/py-netty',
    'license': 'MIT',
    "long_description": long_description,
    "long_description_content_type": 'text/markdown',
    'description': 'TCP framework in flavor of Netty',
    'author' : 'Hao Ruan',
    'author_email': 'ruanhao1116@gmail.com',
    'keywords': ['network', 'tcp', 'non-blocking', 'epoll', 'nio', 'netty'],
    'version': version,
    'packages': find_packages(),
    'install_requires': [],
    'python_requires': ">=3.7, <4",
    'setup_requires': ['wheel'],
    'classifiers': [
        "Intended Audience :: Developers",
        'License :: OSI Approved :: MIT License',
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development :: Libraries",
    ],
}

setup(**config)



import abc
import logging
from .utils import LoggerAdapter
from typing import Union
import socket


logger = LoggerAdapter(logging.getLogger(__name__))


class AbstractChannelHandler(abc.ABC):

    @abc.abstractmethod
    def channel_active(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    @abc.abstractmethod
    def channel_read(self, ctx: 'ChannelHandlerContext', msg: Union[bytes, socket.socket]) -> None:
        pass

    @abc.abstractmethod
    def channel_inactive(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    @abc.abstractmethod
    def channel_registered(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    @abc.abstractmethod
    def channel_unregistered(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    @abc.abstractmethod
    def channel_handshake_complete(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    @abc.abstractmethod
    def exception_caught(self, ctx: 'ChannelHandlerContext', exception: Exception) -> None:
        pass


class ChannelHandlerAdapter(AbstractChannelHandler):

    def channel_active(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    def channel_read(self, ctx: 'ChannelHandlerContext', msg: Union[bytes, socket.socket]) -> None:
        pass

    def channel_inactive(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    def channel_registered(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    def channel_unregistered(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    def channel_handshake_complete(self, ctx: 'ChannelHandlerContext') -> None:
        pass

    def exception_caught(self, ctx: 'ChannelHandlerContext', exception: Exception) -> None:
        pass


NoOpChannelHandler = ChannelHandlerAdapter
DefaultChannelHandler = ChannelHandlerAdapter


class LoggingChannelHandler(AbstractChannelHandler):

    def channel_active(self, ctx: 'ChannelHandlerContext') -> None:
        logger.debug("[Channel Active] %s", ctx.channel())

    def channel_read(self, ctx: 'ChannelHandlerContext', msg: Union[bytes, socket.socket]) -> None:
        logger.debug("[Channel Read] %s", ctx.channel())

    def channel_inactive(self, ctx: 'ChannelHandlerContext') -> None:
        logger.debug("[Channel Inactive] %s", ctx.channel())

    def channel_registered(self, ctx: 'ChannelHandlerContext') -> None:
        logger.debug("[Channel Registered] %s", ctx.channel())

    def channel_unregistered(self, ctx: 'ChannelHandlerContext') -> None:
        logger.debug("[Channel Unregistered] %s", ctx.channel())

    def channel_handshake_complete(self, ctx: 'ChannelHandlerContext') -> None:
        logger.debug("[Channel Handshake Complete] %s", ctx.channel())

    def exception_caught(self, ctx: 'ChannelHandlerContext', exception: Exception) -> None:
        logger.error("[Exception Caught] %s : %s", ctx.channel(), str(exception), exc_info=exception)


class EchoChannelHandler(ChannelHandlerAdapter):

    def __init__(self):
        super().__init__()

    def channel_read(self, ctx: 'ChannelHandlerContext', msg: Union[bytes, socket.socket]) -> None:
        ctx.write(msg)
